/** 
 * SunMicrosystem proprietary information.
 * For use of Adam Megacz only.
 */


package com.sunlabs.archsim.fleet.codeStorage;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Map.Entry;


/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author Igor
 *
 * Storage for Fleet instruction, a singleton.
 * Stored are codebags, accessible by name.
 * Each codebag is a list of instructions.
 * This is an internal structure to be be used by the instruction issue unit
 * and possibly by the instruction issue unit.
 * 
 */

public class CodeStorage {
	
	// the list of code bags that are to be executed first
	public List<CodeBag> bootstrap;
	// code baggs, access is by names
	public Map<String, CodeBag> bags;	
	
	
	private CodeStorage() {
		// make the map for bags
		bags= new HashMap<String, CodeBag>();
		// make the list for the bootstrap
		bootstrap= new LinkedList<CodeBag>();
	}
	
	// instance, used in getInstance
	private static CodeStorage instance=null;
	
	/**
	 * Get an instance of code storage
	 */
	public static CodeStorage getInstance() {
		if (instance == null) {
			instance= new CodeStorage();
		}
		return instance;
	}
	
	/**
	 * Parse a file with FleetCode.
	 * Add the new code bags to the storage.
	 * If duplicate bag names occur, report an error, keep the first bag.
	 * @param fileName file to be parsed.
	 */
	void addBags(String fileName) {
		// TODO implement this
	}

	
	/**
	 * Adds a code bag to CodeStorage.
	 * If the bag has the same name as an existing bag,
	 * the old bag is replaced with the new one.
	 * @param b the CodeBag to be 
	 * @return the replaced code bag, null if no bag was replaced
	 */
	public CodeBag addBag(CodeBag b, boolean boot) {
		// should this codebag be loaded automatically upon startup?
		if (boot) bootstrap.add(b);
		// add the code bag to the repository
		CodeBag oldBag= bags.put(b.getName(), b);
		
		return oldBag;
	}
	
	/**
	 * Shorthand for adding a bag that is not automatically loaded and issued 
	 * at startup/bootstrap. Equivalent of addBag(b, false);
	 * @param b CodeBag to be added
	 * @return null if no code bag was replaced, otherwise a pointer to the replaced bag
	 */
	public CodeBag addBag(CodeBag b) {
		return addBag(b, false);
	} // addBag

	/**
	 * Shorhand for adding a startup bag. Equivalent to addBag(b, true)
	 * @param b CodeBag to be added
	 * @return null if no code bag was replaced, otherwise a pointer to the replaced bag
	 */
	public CodeBag addStartupBag(CodeBag b) {
		return addBag(b, true);
	} // addStartupBag
	
	
	/**
	 * Does the code storage contain a bag with a certain name?
	 * @param name the name of a bag
	 * @return true if the storage contains the bag with the given name, false otherwise
	 */
	public boolean containsBag(String name) {
		return bags.containsKey(name);
	}
	
	/**
	 * Get a codebag given a name.
	 * @param name codebbag name
	 * @return codebag object or null if a codebag does not exist
	 */
	public CodeBag getBag(String name) {
		return bags.get(name);
	}
	
	/**
	 * Used for examining the code storage.
	 * @return Iterator over the bags in the repository. 
	 */
	public Iterator<CodeBag> bagIterator() {
		return bags.values().iterator();
	} // bagIterator
	
	/**
	 * Used for examininig code bag names.
	 * @return Iterator over code bag names.
	 */
	public Iterator<String> bagNameIterator() {
		return bags.keySet().iterator();
	} // bagNameIterator
	
	/**
	 * Get the first code bag to be executed.
	 * Use this until we define a bootstrap mechanism.
	 * @return
	 */
	public Iterator<CodeBag> startupIterator() {
		return bootstrap.iterator();
	} // startupIterator
	
	/**
	 * Empty the code storage.
	 *
	 */
	public void clear() {
		// clear both the bag storage and the boostrap list
		bags.clear();
		bootstrap.clear();
	}
	
	public String toString() {
		String storageString="Code storage:";
		for (Entry<String, CodeBag> e : bags.entrySet()) {
			CodeBag bag= e.getValue();
			String bagName= e.getKey();
			storageString+= "\n";
			if (bootstrap.contains(bag)) {
				storageString+= "INITIAL ";
			}
			storageString = storageString
				+ bagName
				+ ": "
				+ bag.toString();
		}

		return storageString;
	}
	
	
} // CodeStorage


 