package edu.berkeley.sbp.bind;

import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.chr.*;
import edu.berkeley.sbp.bind.*;
import java.util.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.io.*;
import static edu.berkeley.sbp.util.Reflection.*;

public abstract class Bindable implements ToJava {

    public abstract String getSimpleName();
    public abstract String toString();
    public abstract <A extends Annotation> A getAnnotation(Class<A> c);
    public abstract Object impose(Object[] fields);
    public boolean isAnnotationPresent(Class<? extends Annotation> c) { return getAnnotation(c) != null; }

    public abstract Annotation[][] getArgAnnotations();
    public abstract String[]       getArgNames();

    public static Bindable create(Object o) {
        if (o instanceof Class) return new BindableClass((Class)o);
        if (o instanceof Method) return new BindableMethod((Method)o);
        if (o instanceof Constructor) return new BindableConstructor((Constructor)o);
        return null;
    }

    private static class BindableMethod extends Bindable {
        private final Method _method;
        public String toString() { return "BindableMethod["+_method+"]"; }
        public BindableMethod(Method _method) { this._method = _method; }
        public String getSimpleName() { return _method.getName(); }
        public <A extends Annotation> A getAnnotation(Class<A> c) { return _method.getAnnotation(c); }
        public Object impose(Object[] fields) { return Reflection.impose(_method, fields); }
        public Annotation[][] getArgAnnotations() { return _method.getParameterAnnotations(); }
        public String[]       getArgNames() { return new String[_method.getParameterTypes().length]; }
        public void toJava(StringBuffer sb) {
            sb.append("Bindable.create(");
            sb.append(_method.getDeclaringClass().getName().replace('$','.'));
            sb.append(".class.getMethod(\"");
            sb.append(_method.getName());
            sb.append("\", ");
            appendClassArray(sb, _method.getParameterTypes());
            sb.append("))");
        }
    }

    private static class BindableConstructor extends Bindable {
        private final Constructor _constructor;
        public String toString() { return "BindableConstructor["+_constructor+"]"; }
        public BindableConstructor(Constructor _constructor) { this._constructor = _constructor; }
        public String getSimpleName() { return _constructor.getName(); }
        public <A extends Annotation> A getAnnotation(Class<A> c) { return _constructor.getAnnotation(c); }
        public Object impose(Object[] fields) { return Reflection.impose(_constructor, fields); }
        public Annotation[][] getArgAnnotations() { return _constructor.getParameterAnnotations(); }
        public String[]       getArgNames() { return new String[_constructor.getParameterTypes().length]; }
        public void toJava(StringBuffer sb) {
            sb.append("Bindable.create(");
            sb.append(_constructor.getDeclaringClass().getName().replace('$','.'));
            sb.append(".class.getConstructor(");
            appendClassArray(sb, _constructor.getParameterTypes());
            sb.append("))");
        }
    }

    private static class BindableClass extends Bindable {
        private final Class _class;
        public String toString() { return "BindableClass["+_class+"]"; }
        public BindableClass(Class _class) { this._class = _class; }
        public String getSimpleName() { return _class.getSimpleName(); }
        public <A extends Annotation> A getAnnotation(Class<A> c) { return (A)_class.getAnnotation(c); }
        public Object impose(Object[] fields) { return Reflection.impose(_class, fields); }
        public Annotation[][] getArgAnnotations() {
            Field[] fields = _class.getFields();
            Annotation[][] ret = new Annotation[fields.length][];
            for(int i=0; i<fields.length; i++)
                ret[i] = fields[i].getAnnotations();
            return ret;
        }
        public String[]       getArgNames() {
            Field[] fields = _class.getFields();
            String[] ret = new String[fields.length];
            for(int i=0; i<fields.length; i++)
                ret[i] = fields[i].getName();
            return ret;
        }
        public void toJava(StringBuffer sb) {
            sb.append("Bindable.create(");
            sb.append(_class.getName().replace('$','.'));
            sb.append(".class)");
        }
    }


    private static void appendClassArray(StringBuffer sb, Class[] c) {
        sb.append("new Class[] {");
        for(int i=0; i<c.length; i++) {
            sb.append(makeClass(c[i]));
            sb.append(".class");
            if (i<c.length-1)
                sb.append(",");
        }
        sb.append("}");
    }

    private static String makeClass(Class c) {
        if (!c.isArray()) return c.getName().replace('$','.');
        return makeClass(c.getComponentType())+"[]";
    }
}
